@@ -66,7 +66,7 @@ class ParentPollerWindows(Thread):
         assert interrupt_handle or parent_handle
         super().__init__()
         if ctypes is None:
-            msg = "ParentPollerWindows requires ctypes"
+            msg = "ParentPollerWindows requires ctypes"  # type:ignore[unreachable]
             raise ImportError(msg)
         self.daemon = True
         self.interrupt_handle = interrupt_handle
