@@ -5,6 +5,7 @@
 
 import asyncio
 import atexit
+import contextvars
 import io
 import os
 import sys
@@ -12,10 +13,10 @@ import threading
 import traceback
 import warnings
 from binascii import b2a_hex
-from collections import deque
+from collections import defaultdict, deque
 from io import StringIO, TextIOBase
 from threading import local
-from typing import Any, Callable, Deque, Dict, Optional
+from typing import Any, Callable, Optional
 
 import zmq
 from jupyter_client.session import extract_header
@@ -29,6 +30,8 @@ from zmq.eventloop.zmqstream import ZMQStream
 MASTER = 0
 CHILD = 1
 
+PIPE_BUFFER_SIZE = 1000
+
 # -----------------------------------------------------------------------------
 # IO classes
 # -----------------------------------------------------------------------------
@@ -55,6 +58,7 @@ class IOPubThread:
             piped from subprocesses.
         """
         self.socket = socket
+        self._stopped = False
         self.background_socket = BackgroundSocket(self)
         self._master_pid = os.getpid()
         self._pipe_flag = pipe
@@ -62,11 +66,11 @@ class IOPubThread:
         if pipe:
             self._setup_pipe_in()
         self._local = threading.local()
-        self._events: Deque[Callable[..., Any]] = deque()
-        self._event_pipes: Dict[threading.Thread, Any] = {}
+        self._events: deque[Callable[..., Any]] = deque()
+        self._event_pipes: dict[threading.Thread, Any] = {}
         self._event_pipe_gc_lock: threading.Lock = threading.Lock()
         self._event_pipe_gc_seconds: float = 10
-        self._event_pipe_gc_task: Optional[asyncio.Task] = None
+        self._event_pipe_gc_task: Optional[asyncio.Task[Any]] = None
         self._setup_event_pipe()
         self.thread = threading.Thread(target=self._thread_main, name="IOPub")
         self.thread.daemon = True
@@ -81,13 +85,22 @@ class IOPubThread:
             self._event_pipe_gc_task = asyncio.ensure_future(self._run_event_pipe_gc())
 
         self.io_loop.run_sync(_start_event_gc)
-        self.io_loop.start()
+
+        if not self._stopped:
+            # avoid race if stop called before start thread gets here
+            # probably only comes up in tests
+            self.io_loop.start()
+
         if self._event_pipe_gc_task is not None:
             # cancel gc task to avoid pending task warnings
             async def _cancel():
-                self._event_pipe_gc_task.cancel()  # type:ignore
+                self._event_pipe_gc_task.cancel()  # type:ignore[union-attr]
+
+            if not self._stopped:
+                self.io_loop.run_sync(_cancel)
+            else:
+                self._event_pipe_gc_task.cancel()
 
-            self.io_loop.run_sync(_cancel)
         self.io_loop.close(all_fds=True)
 
     def _setup_event_pipe(self):
@@ -204,8 +217,7 @@ class IOPubThread:
         """check for forks, and switch to zmq pipeline if necessary"""
         if not self._pipe_flag or self._is_master_process():
             return MASTER
-        else:
-            return CHILD
+        return CHILD
 
     def start(self):
         """Start the IOPub thread"""
@@ -217,10 +229,16 @@ class IOPubThread:
 
     def stop(self):
         """Stop the IOPub thread"""
+        self._stopped = True
         if not self.thread.is_alive():
             return
         self.io_loop.add_callback(self.io_loop.stop)
-        self.thread.join()
+
+        self.thread.join(timeout=30)
+        if self.thread.is_alive():
+            # avoid infinite hang if stop fails
+            msg = "IOPub thread did not terminate in 30 seconds"
+            raise TimeoutError(msg)
         # close *all* event pipes, created in any thread
         # event pipes can only be used from other threads while self.thread.is_alive()
         # so after thread.join, this should be safe
@@ -301,7 +319,7 @@ class BackgroundSocket:
                 stacklevel=2,
             )
             return getattr(self.io_thread.socket, attr)
-        super().__getattr__(attr)  # type:ignore[misc]
+        return super().__getattr__(attr)  # type:ignore[misc]
 
     def __setattr__(self, attr, value):
         """Set an attribute on the socket."""
@@ -349,13 +367,12 @@ class OutStream(TextIOBase):
         """
         if getattr(self, "_original_stdstream_copy", None) is not None:
             return self._original_stdstream_copy
-        else:
-            msg = "fileno"
-            raise io.UnsupportedOperation(msg)
+        msg = "fileno"
+        raise io.UnsupportedOperation(msg)
 
     def _watch_pipe_fd(self):
         """
-        We've redirected standards steams 0 and 1 into a pipe.
+        We've redirected standards streams 0 and 1 into a pipe.
 
         We need to watch in a thread and redirect them to the right places.
 
@@ -367,11 +384,11 @@ class OutStream(TextIOBase):
         """
 
         try:
-            bts = os.read(self._fid, 1000)
+            bts = os.read(self._fid, PIPE_BUFFER_SIZE)
             while bts and self._should_watch:
-                self.write(bts.decode())
+                self.write(bts.decode(errors="replace"))
                 os.write(self._original_stdstream_copy, bts)
-                bts = os.read(self._fid, 1000)
+                bts = os.read(self._fid, PIPE_BUFFER_SIZE)
         except Exception:
             self._exc = sys.exc_info()
 
@@ -396,7 +413,7 @@ class OutStream(TextIOBase):
         name : str {'stderr', 'stdout'}
             the name of the standard stream to replace
         pipe : object
-            the pip object
+            the pipe object
         echo : bool
             whether to echo output
         watchfd : bool (default, True)
@@ -406,7 +423,7 @@ class OutStream(TextIOBase):
             that will swap the give file descriptor for a pipe, read from the
             pipe, and insert this into the current Stream.
         isatty : bool (default, False)
-            Indication of whether this stream has termimal capabilities (e.g. can handle colors)
+            Indication of whether this stream has terminal capabilities (e.g. can handle colors)
 
         """
         if pipe is not None:
@@ -430,13 +447,19 @@ class OutStream(TextIOBase):
         self.pub_thread = pub_thread
         self.name = name
         self.topic = b"stream." + name.encode()
-        self.parent_header = {}
+        self._parent_header: contextvars.ContextVar[dict[str, Any]] = contextvars.ContextVar(
+            "parent_header"
+        )
+        self._parent_header.set({})
+        self._thread_to_parent = {}
+        self._thread_to_parent_header = {}
+        self._parent_header_global = {}
         self._master_pid = os.getpid()
         self._flush_pending = False
         self._subprocess_flush_pending = False
         self._io_loop = pub_thread.io_loop
         self._buffer_lock = threading.RLock()
-        self._buffer = StringIO()
+        self._buffers = defaultdict(StringIO)
         self.echo = None
         self._isatty = bool(isatty)
         self._should_watch = False
@@ -479,6 +502,30 @@ class OutStream(TextIOBase):
                 msg = "echo argument must be a file-like object"
                 raise ValueError(msg)
 
+    @property
+    def parent_header(self):
+        try:
+            # asyncio-specific
+            return self._parent_header.get()
+        except LookupError:
+            try:
+                # thread-specific
+                identity = threading.current_thread().ident
+                # retrieve the outermost (oldest ancestor,
+                # discounting the kernel thread) thread identity
+                while identity in self._thread_to_parent:
+                    identity = self._thread_to_parent[identity]
+                # use the header of the oldest ancestor
+                return self._thread_to_parent_header[identity]
+            except KeyError:
+                # global (fallback)
+                return self._parent_header_global
+
+    @parent_header.setter
+    def parent_header(self, value):
+        self._parent_header_global = value
+        return self._parent_header.set(value)
+
     def isatty(self):
         """Return a bool indicating whether this is an 'interactive' stream.
 
@@ -513,7 +560,7 @@ class OutStream(TextIOBase):
             self._should_watch = False
             # thread won't wake unless there's something to read
             # writing something after _should_watch will not be echoed
-            os.write(self._original_stdstream_fd, b'\0')
+            os.write(self._original_stdstream_fd, b"\0")
             self.watch_fd_thread.join()
             # restore original FDs
             os.dup2(self._original_stdstream_copy, self._original_stdstream_fd)
@@ -582,28 +629,28 @@ class OutStream(TextIOBase):
                 if self.echo is not sys.__stderr__:
                     print(f"Flush failed: {e}", file=sys.__stderr__)
 
-        data = self._flush_buffer()
-        if data:
-            # FIXME: this disables Session's fork-safe check,
-            # since pub_thread is itself fork-safe.
-            # There should be a better way to do this.
-            self.session.pid = os.getpid()
-            content = {"name": self.name, "text": data}
-            msg = self.session.msg("stream", content, parent=self.parent_header)
+        for parent, data in self._flush_buffers():
+            if data:
+                # FIXME: this disables Session's fork-safe check,
+                # since pub_thread is itself fork-safe.
+                # There should be a better way to do this.
+                self.session.pid = os.getpid()
+                content = {"name": self.name, "text": data}
+                msg = self.session.msg("stream", content, parent=parent)
 
-            # Each transform either returns a new
-            # message or None. If None is returned,
-            # the message has been 'used' and we return.
-            for hook in self._hooks:
-                msg = hook(msg)
-                if msg is None:
-                    return
+                # Each transform either returns a new
+                # message or None. If None is returned,
+                # the message has been 'used' and we return.
+                for hook in self._hooks:
+                    msg = hook(msg)
+                    if msg is None:
+                        return
 
-            self.session.send(
-                self.pub_thread,
-                msg,
-                ident=self.topic,
-            )
+                self.session.send(
+                    self.pub_thread,
+                    msg,
+                    ident=self.topic,
+                )
 
     def write(self, string: str) -> Optional[int]:  # type:ignore[override]
         """Write to current stream after encoding if necessary
@@ -614,9 +661,10 @@ class OutStream(TextIOBase):
             number of items from input parameter written to stream.
 
         """
+        parent = self.parent_header
 
         if not isinstance(string, str):
-            msg = f"write() argument must be str, not {type(string)}"
+            msg = f"write() argument must be str, not {type(string)}"  # type:ignore[unreachable]
             raise TypeError(msg)
 
         if self.echo is not None:
@@ -629,21 +677,21 @@ class OutStream(TextIOBase):
         if self.pub_thread is None:
             msg = "I/O operation on closed file"
             raise ValueError(msg)
+
+        is_child = not self._is_master_process()
+        # only touch the buffer in the IO thread to avoid races
+        with self._buffer_lock:
+            self._buffers[frozenset(parent.items())].write(string)
+        if is_child:
+            # mp.Pool cannot be trusted to flush promptly (or ever),
+            # and this helps.
+            if self._subprocess_flush_pending:
+                return None
+            self._subprocess_flush_pending = True
+            # We can not rely on self._io_loop.call_later from a subprocess
+            self.pub_thread.schedule(self._flush)
         else:
-            is_child = not self._is_master_process()
-            # only touch the buffer in the IO thread to avoid races
-            with self._buffer_lock:
-                self._buffer.write(string)
-            if is_child:
-                # mp.Pool cannot be trusted to flush promptly (or ever),
-                # and this helps.
-                if self._subprocess_flush_pending:
-                    return None
-                self._subprocess_flush_pending = True
-                # We can not rely on self._io_loop.call_later from a subprocess
-                self.pub_thread.schedule(self._flush)
-            else:
-                self._schedule_flush()
+            self._schedule_flush()
 
         return len(string)
 
@@ -652,27 +700,27 @@ class OutStream(TextIOBase):
         if self.pub_thread is None:
             msg = "I/O operation on closed file"
             raise ValueError(msg)
-        else:
-            for string in sequence:
-                self.write(string)
+        for string in sequence:
+            self.write(string)
 
     def writable(self):
         """Test whether the stream is writable."""
         return True
 
-    def _flush_buffer(self):
+    def _flush_buffers(self):
         """clear the current buffer and return the current buffer data."""
-        buf = self._rotate_buffer()
-        data = buf.getvalue()
-        buf.close()
-        return data
+        buffers = self._rotate_buffers()
+        for frozen_parent, buffer in buffers.items():
+            data = buffer.getvalue()
+            buffer.close()
+            yield dict(frozen_parent), data
 
-    def _rotate_buffer(self):
+    def _rotate_buffers(self):
         """Returns the current buffer and replaces it with an empty buffer."""
         with self._buffer_lock:
-            old_buffer = self._buffer
-            self._buffer = StringIO()
-        return old_buffer
+            old_buffers = self._buffers
+            self._buffers = defaultdict(StringIO)
+        return old_buffers
 
     @property
     def _hooks(self):
